/*
 * File: findtext.c
 *
 * Copyright 2003 Sebastian Geerken <s.geerken@ping.de>,
 *                Eric Gaudet <eric@rti-zone.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

/*
 * This module handles selections, as well as activation of links
 * (which is closely related). The Selection structure is associated
 * to GtkDwViewport , and so to a widget tree. Single widgets send
 * "abstract events", by constructing simple iterators (DwIterator),
 * and calling a_Selection_button_press, a_Selection_button_release or
 * a_Selection_button_motion. The arguments in detail:
 *
 *    - DwIterator *it         the iterator pointing on the item under
 *                             the mouse pointer,
 *    - gint char_pos          if this item is a text word, the exact
 *                             character,
 *    - gint link              if this item is associated with a link,
 *                             its number (see DwImage, section
 *                             "signals" for the meaning), otherwise
 *                             -1,
 *    - GdkEventButton *event  the event itself. Only the button is used.
 *
 * todo (most of it just collected here, but they have to be fixed
 * elsewhere):
 *
 * - The selection should be cleared, when the user selects text
 *    somewhere else
 *
 * - Currently, only text is selected. To select things like bullets
 *   (as "*"), images (ALT text), and rulers, the respective widgets
 *   have to be extended by iterators returning exactly one
 *   DW_CONTENT_TEXT.
 *
 * - Container widgets other than DwPage (especially DwTable) do not
 *   process events, which sometimes leads to strange results.
 */

#include "selection.h"
#include "commands.h"
#include <stdio.h>
#include <string.h>
#include <gtk/gtksignal.h>

/*#define DEBUG_LEVEL 1*/
#include "debug.h"

static void Selection_highlight (Selection *selection,
                                 gboolean fl);
static void Selection_copy(Selection *selection);

/* Not defined as static, to avoid compiler warnings. */
char *selection_state_descr[] = { "none", "link_pressed", "selecting",
                                  "selected" };

Selection *a_Selection_new (void)
{
   Selection *selection;

   selection = g_new (Selection, 1);
   selection->from = NULL;
   selection->to = NULL;
   selection->state = SELECTION_NONE;
   selection->link_pressed = -1;
   selection->dclick_callback = NULL;
   return selection;
}


/*
 * Sets a callback function for double click, this is normally
 * the function toggling the full screen mode.
 * (More than one is never needed, so this instead of a signal.)
 */
void a_Selection_set_dclick_callback (Selection *selection,
                                      void (*fn) (gpointer data),
                                      gpointer data)
{
   g_return_if_fail (selection->dclick_callback == NULL);
   selection->dclick_callback = fn;
   selection->callback_data = data;
}


void a_Selection_reset (Selection *selection)
{
   if (selection->from)
      a_Dw_ext_iterator_free(selection->from);
   selection->from = NULL;
   if (selection->to)
      a_Dw_ext_iterator_free(selection->to);
   selection->to = NULL;
   selection->state = SELECTION_NONE;
}


void a_Selection_free (Selection *selection)
{
   a_Selection_reset (selection);
   g_free (selection);
}

gint a_Selection_button_press  (Selection *selection,
                                DwIterator *it,
                                gint char_pos,
                                gint link,
                                GdkEventButton *event,
                                gboolean within_content)
{
   DwWidget *it_widget;
   gboolean ret = FALSE;

   DEBUG_MSG (1, "PRESS(%s, %d, %d): %s -> ...\n",
              a_Dw_iterator_text (it), char_pos, link,
              selection_state_descr[selection->state]);

   Selection_highlight (selection, FALSE);
   a_Selection_reset (selection);

   /* When the user double-clicks on empty parts, call the callback
    * function instead of normal processing. Used for full screen
    * mode. */
   if (!within_content && event != NULL &&
       event->type == GDK_2BUTTON_PRESS && event->button == 1) {
      if (selection->dclick_callback)
         selection->dclick_callback (selection->callback_data);
      return TRUE;
   }

   /* 'it' may be freed by a_Dw_ext_iterator_new() */
   it_widget = it->widget;
   selection->from = a_Dw_ext_iterator_new (it);
   /* a_Dw_ext_iterator_new may return NULL, if the page has no contents. */
   if (selection->from) {
      selection->from_char = char_pos;

      if (link != -1) {
         gtk_signal_emit_by_name (GTK_OBJECT (it_widget), "link_pressed",
                                  link, -1, -1, event);
         selection->state = SELECTION_LINK_PRESSED;
         selection->link_pressed = link;
         ret = TRUE;
      } else {
         if (event && event->button == 1) {
            selection->state = SELECTION_SELECTING;
            selection->to = a_Dw_ext_iterator_clone (selection->from);
            selection->to_char = char_pos;
            ret = TRUE;
         }
      }
   } else
      /* If there is no content */
      ret = FALSE;

   DEBUG_MSG (1, "  ... -> %s, processed = %s\n",
              selection_state_descr[selection->state], ret ? "TRUE" : "FALSE");
   return ret;
}


gint a_Selection_button_release (Selection *selection,
                                 DwIterator *it,
                                 gint char_pos,
                                 gint link,
                                 GdkEventButton *event,
                                 gboolean within_content)
{
   gboolean ret = FALSE;
   DwExtIterator *old_to;

   DEBUG_MSG (1, "RELEASE(%s, %d, %d): %s -> ...\n",
              a_Dw_iterator_text (it), char_pos, link,
              selection_state_descr[selection->state]);

   if (selection->state == SELECTION_LINK_PRESSED) {
      if (link != -1) {
         gtk_signal_emit_by_name (GTK_OBJECT (it->widget), "link_released",
                                  link, -1, -1, event);
         if (link == selection->link_pressed) {
            gtk_signal_emit_by_name (GTK_OBJECT (it->widget), "link_clicked",
                                     link, -1, -1, event);
         }
         selection->state = SELECTION_NONE;
      } else {
         selection->state = SELECTION_SELECTING;
         selection->link_pressed = -1;
      }
   }

   if (selection->state == SELECTION_SELECTING &&
       event && event->button == 1) {
      Selection_highlight (selection, FALSE);
      old_to = selection->to;
      selection->to = a_Dw_ext_iterator_new_variant (old_to, it);
      if (old_to)
         a_Dw_ext_iterator_free(old_to);
      selection->to_char = char_pos;
      if (a_Dw_ext_iterator_compare (selection->from, selection->to) == 0 &&
          selection->from_char == selection->to_char)
         a_Selection_reset (selection);
      else {
         Selection_highlight (selection, TRUE);
         Selection_copy (selection);
         selection->state = SELECTION_SELECTED;
         ret = TRUE;
      }
   }

   DEBUG_MSG (1, "  ... -> %s, processed = %s\n",
              selection_state_descr[selection->state], ret ? "TRUE" : "FALSE");
   return ret;
}


gint a_Selection_button_motion (Selection *selection,
                                DwIterator *it,
                                gint char_pos,
                                gint link,
                                GdkEventButton *event,
                                gboolean within_content)
{
   DwExtIterator *old_to;

   DEBUG_MSG (1, "MOTION (%s, %d, %d): %s -> ...\n",
              a_Dw_iterator_text (it), char_pos, link,
              selection_state_descr[selection->state]);

   if (selection->state == SELECTION_LINK_PRESSED) {
      if (link != selection->link_pressed) {
         selection->state = SELECTION_SELECTING;
         selection->link_pressed = -1;
      }
   }

   if (selection->state == SELECTION_SELECTING) {
      Selection_highlight (selection, FALSE);
      old_to = selection->to;
      selection->to = a_Dw_ext_iterator_new_variant (old_to, it);
      if (old_to)
         a_Dw_ext_iterator_free(old_to);
      selection->to_char = char_pos;
      Selection_highlight (selection, TRUE);
   }

   DEBUG_MSG (1, "  ... -> %s, processed = TRUE\n",
              selection_state_descr[selection->state]);
   return TRUE;
}

static void Selection_highlight (Selection *selection,
                                 gboolean fl)
{
   /* If someone is really interested in optimizing highlighting for
    * large selected areas, unhighlighting old and highlighting new
    * should be replaced by unhighlighting and highlighting the
    * changed parts. */
   DwExtIterator *a, *b, *i;
   gint cmp, a_char, b_char;
   gboolean start;

   if (selection->from && selection->to) {
      DEBUG_MSG (1, "  %shighlighting from %s to %s\n",
                 fl ? "" : "un", a_Dw_ext_iterator_text (selection->from),
                 a_Dw_ext_iterator_text (selection->to));
      cmp = a_Dw_ext_iterator_compare (selection->from, selection->to);
      if (cmp == 0) {
         if (fl) {
            if (selection->from_char < selection->to_char)
               a_Dw_ext_iterator_highlight (selection->from,
                                            selection->from_char,
                                            selection->to_char,
                                            DW_HIGHLIGHT_SELECTION);
            else
               a_Dw_ext_iterator_highlight (selection->from,
                                            selection->to_char,
                                            selection->from_char,
                                            DW_HIGHLIGHT_SELECTION);
         } else
            a_Dw_ext_iterator_unhighlight (selection->from,
                                           DW_HIGHLIGHT_SELECTION);
         return;
      }

      if (cmp < 0) {
         a = selection->from;
         a_char = selection->from_char;
         b = selection->to;
         b_char = selection->to_char;
      } else {
         a = selection->to;
         a_char = selection->to_char;
         b = selection->from;
         b_char = selection->from_char;
      }

      for (i = a_Dw_ext_iterator_clone (a), start = TRUE;
           (cmp = a_Dw_ext_iterator_compare (i, b)) <= 0;
           a_Dw_ext_iterator_next (i), start = FALSE) {
         DEBUG_MSG (1, "    %shighlighting %s\n",
                    fl ? "" : "un", a_Dw_ext_iterator_text (i));
         if (i->content.type == DW_CONTENT_TEXT) {
            if (fl) {
               if (start)
                  a_Dw_ext_iterator_highlight (i, a_char,
                                               strlen (i->content.data.text)
                                               + 1,
                                               DW_HIGHLIGHT_SELECTION);
               else if (cmp == 0)
                  /* the end */
                  a_Dw_ext_iterator_highlight (i, 0, b_char,
                                               DW_HIGHLIGHT_SELECTION);
               else
                  a_Dw_ext_iterator_highlight (i, 0,
                                               strlen (i->content.data.text)
                                               + 1,
                                               DW_HIGHLIGHT_SELECTION);
            } else
               a_Dw_ext_iterator_unhighlight (i, DW_HIGHLIGHT_SELECTION);
         }
      }
      a_Dw_ext_iterator_free (i);
   }
}


static void Selection_copy(Selection *selection)
{
   DwIterator *si;
   DwExtIterator *a, *b, *i;
   gint cmp, a_char, b_char;
   gboolean start;
   gchar *sel = g_strdup("");
   gchar *txt;
   gchar *tmp;

   if (selection->from && selection->to) {
      cmp = a_Dw_ext_iterator_compare (selection->from, selection->to);
      if (cmp == 0) {
         if (selection->from->content.type == DW_CONTENT_TEXT) {
            si = selection->from->stack[selection->from->stack_top];
            if (selection->from_char < selection->to_char) {
               txt = g_strndup(si->content.data.text + selection->from_char,
                               selection->to_char - selection->from_char);
            } else {
               txt = g_strndup(si->content.data.text + selection->to_char,
                               selection->from_char - selection->to_char);
            }
            tmp = g_strconcat(sel,txt,NULL);
            g_free(sel);
            g_free(txt);
            sel = tmp;
         }
      } else {
         if (cmp < 0) {
            a = selection->from;
            a_char = selection->from_char;
            b = selection->to;
            b_char = selection->to_char;
         } else {
            a = selection->to;
            a_char = selection->to_char;
            b = selection->from;
            b_char = selection->from_char;
         }

         for (i = a_Dw_ext_iterator_clone (a), start = TRUE;
              (cmp = a_Dw_ext_iterator_compare (i, b)) <= 0;
              a_Dw_ext_iterator_next (i), start = FALSE) {
            si = i->stack[i->stack_top];
            switch (si->content.type) {
            case DW_CONTENT_TEXT:
               if (start)
                  txt = g_strndup(si->content.data.text + a_char,
                                  strlen (i->content.data.text) - a_char);
               else if (cmp == 0)
                  /* the end */
                  txt = g_strndup(si->content.data.text,
                                  b_char);
               else
                  txt = g_strndup(si->content.data.text,
                                  strlen (i->content.data.text));
               tmp = g_strconcat(sel,txt,NULL);
               g_free(sel);
               g_free(txt);
               sel = tmp;
               if (si->content.space && cmp != 0) {
                  tmp = g_strconcat(sel," ",NULL);
                  g_free(sel);
                  sel = tmp;
               }
               break;

            case DW_CONTENT_BREAK:
               if (si->content.data.break_space > 0)
                  tmp = g_strconcat(sel,"\n\n",NULL);
               else
                  tmp = g_strconcat(sel,"\n",NULL);
               g_free(sel);
               sel = tmp;
               break;
            default:
               /* Make pedantic compilers happy. Especially
                * DW_CONTENT_WIDGET is never returned by a DwExtIterator. */
               break;
            }
         }
         a_Dw_ext_iterator_free (i);
      }
   }
   a_Commands_set_selection(sel);
   g_free(sel);
}
